# Copyright (c) 2019 Qumulo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# -----------------------------------------------------------------------------
# QCluster.py
#
# This class contains all the code necessary to deal with a Qumulo Cluster;
# including Clearing the Cache, Getting the Cluster Type, Getting the Cluster IPs,
# and finding out if the Cluster is Rebuilding.


# Import python libraries

import sys
import re
import os
import arrow
import time

from contextlib import contextmanager

# Import Qumulo REST Libraries

import qumulo.lib.auth
import qumulo.lib.opts
import qumulo.lib.request as request
import qumulo.rest
from qumulo.lib.uri import UriBuilder
from qumulo.rest_client import RestClient

# Import Logger

from QumuloUtils.qlogger import Logger

#
# QCluster Class
#
# This class contains all the code necessary to deal with a Qumulo Cluster;
# including finding files based upon age, file type, and/or file size.


class QCluster (object):

    def __init__ (self, qhost, qlogin = "admin", qpasswd = "admin", qport = 8000, logger = None):

        self.logger = logger
        self.qhost = qhost
        self.qport = int (qport)
        self.qlogin = qlogin
        self.qpasswd = qpasswd
        self.connection = None
        self.credentials = None
        self.resultsdir = None
        self.saved_files = []
        self.bucket_number = 1
        self.total_dirs = 0
        self.found_dirs = 0
        self.total_files = 0
        self.found_files = 0
        self.type_sizes = []
        self.type_counts = []
        self.first_scan = True
        self.nodeinfo = None

        # Create a connection to the cluster

        try:
            self.connection = RestClient (self.qhost, self.qport)

            # Login and store credentials

            self.credentials = self.connection.login (self.qlogin, self.qpasswd)
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to login to Qumulo cluster. Error is %s', excpt)

            raise

        return

    # Clear the cache on the cluster

    def ClearCache (self):

        # We are building this "by hand" as we don't want to have
        # the users of these scripts load the internal qumulo libraries.

        method = "POST"
        uri = str (UriBuilder (path = "/v1/debug/cache/clear"))

        try:
            request.rest_request (self.connection, self.credentials, method, uri)
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Exception on clearing cache. Error is %s', excpt)

            raise

        return

    # Get some info from the cluster for logging

    def ClusterInfo (self):

        # First, get the model number from the nodes. We will store the
        # node_info structure if they need it for later.

        try:
            self.nodeinfo = self.connection.cluster.list_nodes ()
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to list_nodes in cluster. Error is %s', excpt)

            raise

        # Get the number of nodes

        node_count = len (self.nodeinfo)

        # Get the model number from the nodes. "There can be only one!"

        model = self.nodeinfo[0]['model_number']

        # Get the revision for the cluster

        try:
            versinfo = self.connection.version.version ()
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to get version from cluster. Error is %s', excpt)

            raise

        # Get the version of software from the cluster.

        version = versinfo['revision_id']

        # Get the name of the cluster

        try:
            nameinfo = self.connection.cluster.get_cluster_conf ()
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to get name of the cluster. Error is %s', excpt)

            raise

        # Get the name of the cluster

        clustername = nameinfo['cluster_name']

        return (model, version, clustername, node_count)

    # Get the connection for making "raw" calls

    def GetConnection (self):

        return self.connection

    # Find out what IP addresses are in use by the cluster.
    # We will look at both the "management" and the floating
    # IP's. If they have floating IP's, those will always
    # be used over the "management" IP's.

    def ClusterIPs (self):

        # Get the status of the network. This will return both the
        # "management" and the floating IP's. The floating IP's will
        # have an array length of 0 if there are none. So, this is easy!

        try:
            ifinfo = self.connection.network.list_interfaces ()
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to get network interfaces from cluster. Error is %s',
                                   excpt)
            raise

        try:
            netinfo = self.connection.network.list_network_status_v2 (ifinfo[0]['id'])
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to get network information from cluster. Error is %s',
                                   excpt)
            raise

        ips = []

        for info in netinfo:
            netstatus = info['network_statuses']

            if len (netstatus[0]['floating_addresses']) > 0:
                for floats in netstatus[0]['floating_addresses']:
                    ips.append (floats)
            else:
                ips.append (netstatus[0]['address'])

            # Sort the IPs and then return it

        ips.sort()

        return ips

    # Find out if the cluster is doing a rebuild. This is only allowed if
    # the benchmarker wants this (set via the configuration file). Otherwise,
    # this is an error condition.

    def ClusterRebuilding (self):

        try:
            rebuildinfo = self.connection.cluster.get_restriper_status ()
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to get restriper status from cluster. Error is %s',
                                   excpt)

            raise

        status = rebuildinfo['status']

        if status == "NOT_RUNNING":
            return False

        return True

    # Cycle through a named set of snapshots and delete the oldest greater than the limit

    def CycleSnapshot (self, snap_list, limit = None):

        # If there is a limit, then delete all older snapshots greater than the limit

        if limit is not None:
            our_limit = 1
        else:
            our_limit = limit

        # If we don't have enough snapshots to match the limit, then move on

        if len (snap_list) <= our_limit:
            return

        # Delete all snapshots greater than the limit

        num_del = len (snap_list) - our_limit
        if num_del > 0:
            for ind in range (num_del):
                snap_id = snap_list(ind)['id']

                try:
                    self.connection.snapshot.delete_snapshot (snap_id)
                except Exception, excpt:
                    self.logger.error ("Snapshot id %s was incorrectly deleted.", snap_id)

        return

    # Delete all of the old snapshots but one...

    def RemoveOldSnapshot (self, name):

        # First, get all of the snapshots that match our name

        snap_list = self.FindSnapshot (name)

        # Now, delete all of them except for the last one

        for indx in range (0, len (snap_list) - 1):
            try:
                self.connection.snapshot.delete_snapshot (snap_list[indx]["id"])
            except Exception, excpt:
                self.logger.error ("Snapshot %d was already deleted.", snap_list[indx]["id"])

        return

    # Create a new snapshot based upon the snapshot name

    def CreateSnapshot (self, name, path):

        try:
            self.connection.snapshot.create_snapshot (name = name, path = path)
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to create a snapshot. Error is %s', excpt)

        return

    # Compare two snapshots and get the delta differences

    def CompareSnapshot (self, name):

        # First, get a list of the snapshots that match our name...

        snap_list = self.FindSnapshot (name)

        # We only need the last two of them...

        created_list = []
        modified_list = []
        deleted_list = []
        compare_list = []

        if len (snap_list) >= 2:
            range_start = len (snap_list) - 2
            for indx in range (range_start, len (snap_list)):
                compare_list.append (snap_list[indx]["id"])

            # Get the differences between the two snapshots

            old_id = compare_list[0]
            young_id = compare_list[1]

            changed_list = []

            # Get all of the changes

            next_page = "first"
            while next_page is not None:
                try:
                    if next_page == "first":
                        snap_entries = self.connection.snapshot.get_snapshot_tree_diff (young_id, old_id)
                    else:
                        snap_entries = self.connection.request ("GET", snap_entries["paging"]["next"])
                except Exception, excpt:
                    self.logger.debug ("Error retrieving the snapshot diffs with old %d and young %d ids.",
                                       old_id, young_id)

                next_page = snap_entries["paging"]["next"]

                # Iterate through the list and get all of the created, modified, and deleted
            
                if len (snap_entries["entries"]) > 0:
                    (c_list, m_list, d_list) = self.__ret_snap_diffs (snap_entries["entries"])

                created_list.append (c_list)
                modified_list.append (m_list)
                deleted_list.append (d_list)

        return (created_list, modified_list, deleted_list)

    # Iterate through a snapshot differences list and get all of the created, modified, and deleted

    def __ret_snap_diffs (self, snap_diffs):

        created_list = []
        modified_list = []
        deleted_list = []

        # Find out how many entries that we have and get all of the differences

        for indx in range (len (snap_diffs)):
            if snap_diffs[indx]["op"] == "CREATE":
                created_list.append (snap_diffs[indx])
            elif snap_diffs[indx]["op"] == "MODIFY":
                modified_list.append (snap_diffs[indx])
            else:
                deleted_list.append (snap_diffs[indx])

        return (created_list, modified_list, deleted_list)

    # Find a group of snapshots based upon the snapshot name

    def FindSnapshot (self, name):

        try:
            snap_list = self.connection.snapshot.list_snapshots ()
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to list snapshots. Error is %s', excpt)
            return

        # Build a list of the snapshots that match the name

        valid_snaps = []
        snaps = snap_list["entries"]

        for snap in range (len (snaps)):
            if snaps[snap]["name"] == name:
                valid_snaps.append (snaps[snap])

        return valid_snaps

    # Create a directory

    def CreateDir (self, name, dir_path):

        try:
            self.connection.fs.create_directory (name, dir_path)
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to create directory %s in path %s. Error is %s',
                                   name, dir_path, excpt)

            raise

        return

    # Create a file...
    # You cannot write a file without first creating it.

    def CreateFile (self, name, dataBuffer):

        try:
            dirinfo = self.connection.fs.create_file (name, dataBuffer)
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to create file %s using API. Error is %s',
                                   name, excpt)

            raise

        return dirinfo['id']

    # Write a file...
    # The file must first exist before you can write into it
    # Also, the file command actually reads from a file in your filesystem
    # in order to copy the data.

    def WriteFile (self, id, procFile):

        try:
            self.connection.fs.write_file (procFile, id_=id)
        except Exception, excpt:
            if self.logger is not None:
                self.logger.error ('Unable to write api file. Error is %s', excpt)

            raise

        return




# Copyright (c) 2015 Qumulo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# -----------------------------------------------------------------------------
# TimeCall.py
#
# Class that handles timing of operations. You can frame any operation
# with a "timing" and then report on it anytime during the execution
# of your routine.

'''
== Import

from QumuloUtils import TimeCall

== Invocation

@TimeCall
def SomeRoutineThatYouHaveWritten:

Put the @TimeCall just before the definition of the routine that you wish to time.
That is all that you have to do to capture how long each routine takes to execute.

== Reporting

When you wish to output the counts and timing, you only have to issue a print or some other
statement where you pass in one of the TimeCall routines.

print ("Counts %s" % TimeCall.counts())
print ("Minimums %s" % TimeCall.minimums())
print ("Maximums %s" % TimeCall.maximums())
print ("Averages %s" % TimeCall.averages())

'''

# Import python libraries

import sys
import os
import time
import json
import functools

# TimeCall Class

class TimeCall (object):

   __instances = {}

   def __init__(self, f):
      self.__f = f
      self.__numcalls = 0
      self.__min = 0
      self.__max = 0
      self.__total = 0
      TimeCall.__instances[f] = self

   def __call__(self, *args):
      self.__numcalls += 1
      ts = time.time ()
      result = self.__f (*args)
      te = time.time ()

      # Compute time elapsed and then return

      telapsed = te - ts
      self.__total += telapsed
      self.__min = telapsed if self.__min == 0 else self.__min
      self.__min = telapsed if telapsed < self.__min else self.__min
      self.__max = telapsed if telapsed > self.__max else self.__max

      return result

   def __get__ (self, instance, instancetype):
       return functools.partial (self.__call__, instance)

   def count (self):
      # Return the number of times the function f was called.
      return TimeCall.__instances[self.__f].__numcalls

   def min (self):
       # Return the minimum amount of time a call took
       return "%.5gms" % (TimeCall.__instances[self.__f].__min * 1000)

   def max (self):
       # Return the maximum amount of time a call took
       return "%.5gms" % (TimeCall.__instances[self.__f].__max * 1000)

   def avg (self):
       # Return the average amount of time a call took
       # It is possible that it was never called, so take that into account.

       if TimeCall.__instances[self.__f].__numcalls == 0:
           result = 0
       else:
           result = ((TimeCall.__instances[self.__f].__total / \
                          TimeCall.__instances[self.__f].__numcalls) * 1000)

       return "%.5gms" % result

   # Return a dict of counts for all registered functions."

   @staticmethod
   def counts ():
      return json.dumps (dict ([(f.__name__, \
                                     TimeCall.__instances[f].__numcalls) \
                                   for f in TimeCall.__instances]))

   # Return a dict of minimums for all registered functions."

   @staticmethod
   def minimums ():
       return json.dumps (dict ([(f.__name__, \
                                      TimeCall.__instances[f].min ()) \
                                     for f in TimeCall.__instances]))

   # Return a dict of maximums for all registered functions."

   @staticmethod
   def maximums ():
       return json.dumps (dict ([(f.__name__, \
                                      TimeCall.__instances[f].max ()) \
                                     for f in TimeCall.__instances]))

   # Return a dict of averages for all registered functions."

   @staticmethod
   def averages ():
       return json.dumps (dict ([(f.__name__, \
                                      TimeCall.__instances[f].avg ()) \
                                     for f in TimeCall.__instances]))




# Copyright (c) 2019 Qumulo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# -----------------------------------------------------------------------------
# QAnalytics.py
#
# This class contains all the code necessary to get some data analytics from a
# Qumulo cluster using the API.

# Import python libraries

import os
import sys
import glob
import time
import datetime
import multiprocessing
import Queue
import logging

# Import Qumulo Libraries

from QumuloUtils.qcluster import *
from QumuloUtils.qlogger import Logger

if sys.version_info[0] < 3:
    reload(sys)
    sys.setdefaultencoding('utf8')

# It is ALMOST impossible to debug spawned processes as they run in a separate context.
# The following methodology works perfectly. HOWEVER, a word of caution!!
#
# Make sure that you set your "thread" count to 1. This will work if you spawn hundreds
# of "threads", but the debugging output would be unworkable (to put it mildly).

import pdb

class ForkedPdb(pdb.Pdb):
    """A Pdb subclass that may be used
    from a forked multiprocessing child

    """
    def interaction(self, *args, **kwargs):
        _stdin = sys.stdin
        try:
            sys.stdin = open('/dev/stdin')
            pdb.Pdb.interaction(self, *args, **kwargs)
        finally:
            sys.stdin = _stdin

# Leave the above code in permanently. It is NOT active until you place a 
# ForkedPdb().set_trace() in the code at the place you wish to start debugging.

# The following class allows us to build a counter for a value that
# works with multiprocessing

class Counter(object):
    def __init__(self, initval = 0):
        self.val = multiprocessing.Value('i', initval)
        self.lock = multiprocessing.Lock()

    def increment(self):
        with self.lock:
            self.val.value += 1

    def decrement(self):
        with self.lock:
            self.val.value -= 1

    def value(self):
        with self.lock:
            return self.val.value

#
# This class will handle all of the work for processes 2 thru N
# In the multiprocessing model (versus the Threading model), this is
# not supposed to be a class. I find that handling this as a class is
# far easier to debug as I can separate the functionality. However,
# this class needs to be instantiated by a function (which is the
# starting point for each process.

class WorkerProc (multiprocessing.Process):
    def __init__ (self, dir_q, callback = None, call_info = None,
                  host = None, login = None, passwd = None,
                  port = None, logger = None, count_in_q = None,
                  count_finished = None):

        super (WorkerProc, self).__init__ ()

        self.dir_q = dir_q
        self.count_in_q = count_in_q
        self.count_finished = count_finished
        self.qcluster = None
        self._callback = None
        self.proc_name = multiprocessing.current_process().name

        # The callback needs to be instantiated per thread. So, it is done here.

        if callback is not None:
            if call_info is not None:
                self._callback = callback (call_info)
            else:
                self._callback = callback ()

        # The normal logger will not work in a spawned process. So, we will
        # setup to ONLY output errors with the logger. If you set it to DEBUG,
        # then it will give you a whole lot more during any code testing.

        logger = multiprocessing.log_to_stderr ()
        logger.setLevel (logging.ERROR)
        logger.debug ("%s - Init done for Worker Process of QAnalytics.", self.proc_name)

        # This is a forked process and nothing survived. We need to re-instantiate
        # everything. Start by making a connection to the cluster.

        self.qcluster = QCluster (host, login, passwd, port, logger)
        return

    # Take a request from the queue and call the routine that will do the work

    def run (self):

        super (WorkerProc, self).run ()

        # As long as we weren't asked to stop, try to take new tasks from the queue.
        # The tasks are taken with a blocking "get", so no CPU cycles are wasted
        # while waiting. Also, "get is given a timeout, so stoprequest is always
        # checked, even if there's nothing in the queue.

        nowork_count = 0

        while True:
            try:
                entry = self.dir_q.get (True, 0.05)
                self.count_in_q.decrement ()

                self.list_dir (self.qcluster, self._callback, entry)

                self.count_finished.increment ()
                self.dir_q.task_done ()
            except Queue.Empty:

                # Let's terminate if nothing appears in the queue after 50 empty attempts
                # Since our queue empty timeout is 1/10 of a second, this should = 5 seconds

                nowork_count += 1
                if nowork_count == 50:
                    break

                continue

        # The queue has been empty for seconds. Let's leave after calling the helper
        # TearDown function.

        if self._callback is not None:
            self._callback.TearDown ()

        return

    #
    # List each entry in the directory

    def list_dir (self, rc, callback, d):

        # Get the directory entries.. If we have a "next page", then retrieve that instead.
        
        qconn = rc.GetConnection ()

        next_page = "first"
        while next_page != "":
            if next_page == "first":
                r = qconn.fs.read_directory(path=d["path"], page_size=1000)
            else:
                r = qconn.request("GET", r['paging']['next'])
            next_page = r['paging']['next']

            # Process each entry returned. If a file, then process through a "callback routine" if
            # it exists. If a directory, then put it on the queue for further processing by another
            # thread.

            for ent in r["files"]:

                # Does the callback exist? If so, then call it and pass in the file ref id

#                 ForkedPdb().set_trace ()

                # If this is a file, then process it...

                if ent["type"] == "FS_FILE_TYPE_FILE":

                    if callback is not None:
                        callback.process ("cfile", {"process-entry": ent})

                # If a directory, then we need to do some further processing

                if ent["type"] == "FS_FILE_TYPE_DIRECTORY" and int(ent["child_count"]) > 0:

                    # Before we put it in the queue for processing by another thread, we
                    # will add the directory entry. In order to do that, we will need to
                    # get the directory aggregates so that we have an accurate file and subdir
                    # count

                    # Now put it on the queue for processing by another thread
                    
                    self.dir_q.put ({"path": d["path"] + ent["name"] + "/"})
                    self.count_in_q.increment ()

                    # Call the callback object to process the directory entry (if it exists)

                    if callback is not None:
                        _id = ent["id"]
                        cur_dir = qconn.fs.read_dir_aggregates (id_ = _id, max_entries = 0)

#                        ForkedPdb().set_trace ()

                        callback.process ("cdir", {"process-entry": cur_dir, "entry": ent})

#
# QAnalytics Class
#
# This class inherits the QCluster class from QumuloUtils in order to work with a
# Qumulo Cluster API. 
#
# This class will perform either a full or partial scan of a Qumulo filesystem using
# the API in order to find certain user data patterns. It will optionally call a routine
# (callback provided by customer) that will provide additional individual file processing.

class QAnalytics (QCluster):

    def __init__ (self, qhost, qlogin = "admin", qpasswd = "admin", qport = 8000, 
                  logger = None, callback = None, elasticinfo = None, 
                  start_path = None, num_threads = None):

        super (QAnalytics, self).__init__ (qhost = qhost, qlogin = qlogin, qpasswd = qpasswd,
                                           qport = qport, logger = logger)

        self.start_path = []

        for indx in range (len (start_path)):
            self.start_path.append (start_path[indx])

            if self.start_path[indx][-1] != "/":
                self.start_path[indx] = self.start_path[indx] + "/"

                if logger is not None:
                    self.logger.debug ("File analysis on Qumulo cluster %s starting at path %s" % 
                                       (qhost, self.start_path[indx]))

        # Create Queues and Multiprocessing Pool

        if num_threads == None:
            self.num_threads = 1
        else:
            self.num_threads = num_threads

        self.dir_q = multiprocessing.JoinableQueue ()
        self.count_in_q = Counter ()
        self.count_finished = Counter ()

        # Create workers based upon number of threads
        
        if logger is not None:
            self.logger.debug ("Creating %d worker processes to help with finding the data." % 
                              (self.num_threads))

        # Get a list of all the IP's for the cluster. In order to get a fair distribution
        # for all of the threads (processes) across the nodes, we cannot trust that the
        # users have given us a cluster name that is round-robin'd in the DNS.

        IPs = self.ClusterIPs ()

        # Spawn all of the threads (processes)

        self.processes = []

        cur_ip = 0

        for i in range (self.num_threads):
            p = WorkerProc (dir_q = self.dir_q,
                            callback = callback,
                            call_info = elasticinfo,
                            host = IPs[cur_ip],
                            login = qlogin,
                            passwd = qpasswd,
                            port = qport,
                            logger = logger,
                            count_in_q = self.count_in_q,
                            count_finished = self.count_finished)
            self.processes.append (p)

            # Make sure we select the next IP

            cur_ip += 1
            if cur_ip == len (IPs) - 1:
                cur_ip = 0

        # Start all the worker procs

        for p in self.processes:
            p.start ()

        return

    # TearDown the threads (processes). This is only called if somebody hit "ctrl-C"

    def TearDown (self):

        if self.logger is not None:
            self.logger.debug ("TearDown in QAnalytics. Killing all threads.")

        for p in self.processes:
            p.terminate ()

        for p in self.processes:
            p.join ()

        return

    # Routine to do a incremental scan. This compares two snapshots and gets a list
    # of created, modified, and deleted files.

    def IncScan (self, snap_name):

        created, modified, deleted = self.CompareSnapshots (snap_name)

        return

    # Routine to do a full scan from a "start_path"

    def FullScan (self):

        # Read the aggregates from the start_path... 

        for indx in range (len (self.start_path)):
            root = self.connection.fs.read_dir_aggregates (path = self.start_path[indx], max_entries = 0)

            if self.logger is not None:
                self.logger.debug ("Directories to analyze: %12s" % 
                                   "{:,}".format(int(root["total_directories"])))
                self.logger.debug ("      Files to analyze: %12s" % 
                                   "{:,}".format(int(root["total_files"])))

                # Add this directory to a queue to process

                self.dir_q.put ({"path": self.start_path[indx]})
                self.count_in_q.increment ()

        time.sleep (1) # wait a bit for the queue to get build up.

        wait_count = 0
        while not self.dir_q.empty ():
            wait_count += 1
            if (wait_count % 50) == 0: # show status every ~5 seconds
                if self.logger is not None:
                    self.logger.debug ("Processed %s entries. Queue length: %s" % \
                                           (self.count_finished.value (), 
                                            self.count_in_q.value ()))
            time.sleep(1)

        for p in self.processes:
            p.join()

        if self.logger is not None:
            self.logger.debug ("Processed %s entries." % self.count_finished.value ())
            self.logger.debug ("Done with QAnalytics.")



# Copyright (c) 2016 Qumulo, Inc. All rights reserved.
#
# NOTICE: All information and intellectual property contained herein is the
# confidential property of Qumulo, Inc. Reproduction or dissemination of the
# information or intellectual property contained herein is strictly forbidden,
# unless separate prior written permission has been obtained from Qumulo, Inc.

# Copyright (c) 2016 Qumulo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

from setuptools import setup

setup(
    name = "QumuloUtils",
    version = "2.0.7",
    packages = ['QumuloUtils'],
    author = "Michael Kade",
    author_email = "mkade@qumulo.com",
    keywords = "Qumulo Core API Helpers",
    license='LICENSE.txt',
    description = "Utility API routines to assist with the Qumulo clustered filesystem.",
    long_description=open('README.txt').read(),
    url = "http://www.qumulo.com/",
    install_requires=[
        "enum",
        "arrow",
        "logging",
        "python-dateutil",
        "qumulo-api",
        "multiprocessing",
    ],
)



# Copyright (c) 2015 Qumulo, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
# -----------------------------------------------------------------------------
# QumuloLogger.py
#
# Class that assists with logging errors to stdout, stderr, and/or a file.

# Import python libraries

import sys
import re
import os
import logging
import time
from enum import Enum

#
# Build an enumeration for the logging level

class Level (Enum):
    CRITICAL = logging.CRITICAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

#
# Logger Class
#
# This class deals with using the "logging" function of python. Although it could
# be called directly within any class referencing this library, it was thought that
# encapsulating it here would make it easier to deal with. Especially for those who
# have never used logging before.

class Logger (object):

    def __init__ (self, benchName = 'Logger', logLevel = Level.DEBUG):

        # Turn on logger and set level based upon "logLevel" argument

        self.logger = logging.getLogger (benchName)
        self.logger.setLevel (Level.DEBUG)
        self.logger.propagate = 1

        # Create several handlers for the logging service.
        # Set the stream handler to stderr

        logh = logging.StreamHandler ()
        logh.setLevel (logLevel)
        self.logger.addHandler (logh)

    # Change loglevel

    def setLevel (self, logLevel):

        self.logger.setLevel (logLevel)

    # Create log files

    def logFiles (self, baseDir = "/tmp", model = "NoModel", version = "noVersion"):

        resultsDir = self.__buildResultsDir (baseDir, model, version)

        infoh = logging.FileHandler (resultsDir + "/info.log")
        infoh.setLevel (Level.INFO)
        self.logger.addHandler (infoh)

        errorh = logging.FileHandler (resultsDir + "/error.log")
        errorh.setLevel (Level.ERROR)
        self.logger.addHandler (errorh)

        debugh = logging.FileHandler (resultsDir + "/debug.log")
        debugh.setLevel (Level.DEBUG)
        self.logger.addHandler (debugh)

        return resultsDir

    # Write to logging using the "info" level

    def info (self, msg, *vargs):

        self.logger.info (msg, *vargs)

    # Write to logging using the "error" level

    def error (self, msg, *vargs):

        self.logger.error (msg, *vargs)

    # Write to logging using the "debug" level

    def debug (self, msg, *vargs):

        self.logger.debug (msg, *vargs)

    # Build a results directory to store the program results. This will be
    # based upon several items: model and version of cluster, current date
    # and time.

    def __buildResultsDir (self, baseDir, model, version):

        # Get version... It is the third item in the array (2 starting from 0)

        info = version.rsplit (' ')
        versinfo = info[2]

        # Get the date and time for format it

        dateinfo = time.strftime ("%Y%m%d%H%M%S")

        # Build the results directory such:
        #
        # resultsdir / model / version / datetime

        resultsinfo = baseDir + "/" + model + "/" + versinfo + "/" + dateinfo

        # We have to make sure that the directory is created before
        # we can use it. The assumption is that the directory does
        # not exist.

        try:
            os.makedirs (resultsinfo, 0755)
        except Exception, excpt:
            self.logger.error ('Error on makedirs = %s', excpt)
            sys.exit (1)

        return resultsinfo
